@@ -207,10 +207,13 @@ impl IndexSchema {
     /// Return an error if the given key does not match this schema.
     #[inline]
     pub fn validate_key(&self, key: Key) -> TCResult<Key> {
-        if key.len() != self.key.len() {
+        let key_len = key.len();
+        if key_len != self.key.len() {
             return Err(TCError::unsupported(format!(
-                "invalid key {} for schema {}",
+                "key {} has {} columns, but the table key has {}: {}",
                 Tuple::from(key),
+                key_len,
+                self.key.len(),
                 self
             )));
         }
